"use strict";
import * as pbjs from "protobuf";
import ServiceNameTemplate from "ServiceNameTemplate";
import PBMsgDictTemplate from "PBMsgDictTemplate";
import CookieForPath from "CookieForPath";

const fs = nodeRequire("fs");
const path = nodeRequire("path");
const http = nodeRequire("http");
const process = nodeRequire('child_process');
const clipboard = nodeRequire('electron').clipboard;

const $g = (id) => { return <HTMLInputElement>document.getElementById(id) };
/**
 * 输出日志
 */
function log(msg: string) {
	let txtLog = $g("txtLog");
	if (txtLog) {
		txtLog.innerHTML += `[${new Date().format("HH:mm:ss")}] ${msg} <br/>`;
	}
}

/**
 * 输出错误
 */
function error(msg: string) {
	log(`<font color="#f00">${msg}</font>`);
}

const Options = {
	/**
	 * 服务名称
	 */
	ServiceName: "(service)",
	/**
	 * 客户端模块
	 */
	ClientModule: "(cmodule)",
	/**
	 * 客户端路径，和前缀路径拼接得到文件生成路径地址
	 */
	ClientPath: "(cpath)",
	/**
	 * 统一协议号请求间隔时间
	 */
	Limit: "(limit)",
}

interface GlobalCfg 
{
	/**
	 * 客户端项目路径
	 */
	cprefix: string;
	/**
	 * 生成时，使用的wiki地址
	 */
	url?: string;
	/**
	 * ServiceName常量文件的文件相对路径
	 */
	serviceName?: string;
	/**
	 * PB消息字典的文件相对路径
	 */
	pBMsgDictName?: string;
}

const TYPE_DOUBLE: number = 1;
const TYPE_FLOAT: number = 2;
const TYPE_INT64: number = 3;
const TYPE_UINT64: number = 4;
const TYPE_INT32: number = 5;
const TYPE_FIXED64: number = 6;
const TYPE_FIXED32: number = 7;
const TYPE_BOOL: number = 8;
const TYPE_STRING: number = 9;
const TYPE_GROUP: number = 10;
const TYPE_MESSAGE: number = 11;
const TYPE_BYTES: number = 12;
const TYPE_UINT32: number = 13;
const TYPE_ENUM: number = 14;
const TYPE_SFIXED32: number = 15;
const TYPE_SFIXED64: number = 16;
const TYPE_SINT32: number = 17;
const TYPE_SINT64: number = 18;

const type2number = (function () 
{
	let t = {};
	t["double"] = TYPE_DOUBLE;
	t["float"] = TYPE_FLOAT;
	t["int64"] = TYPE_INT64;
	t["uint64"] = TYPE_UINT64;
	t["int32"] = TYPE_INT32;
	t["fixed64"] = TYPE_FIXED64;
	t["fixed32"] = TYPE_FIXED32;
	t["bool"] = TYPE_BOOL;
	t["string"] = TYPE_STRING;
	t["group"] = TYPE_GROUP;
	t["message"] = TYPE_MESSAGE;
	t["bytes"] = TYPE_BYTES;
	t["uint32"] = TYPE_UINT32;
	t["enum"] = TYPE_ENUM;
	t["sfixed32"] = TYPE_SFIXED32;
	t["sfixed64"] = TYPE_SFIXED64;
	t["sint32"] = TYPE_SINT32;
	t["sint64"] = TYPE_SINT64;
	return t;
})();

const rule2number = 
{
	"optional": 1,
	"required": 2,
	"repeated": 3
};

const Type_Null = 0;
const Type_Boolean = 1;
const Type_String = 2;
const Type_Bytes = 4;
const Type_Double = 5;
const Type_Int32 = 6;
const Type_Uint32 = 7;

function field2type(field: pbjs.ProtoField): [string, boolean, string | number, boolean] 
{
	let type = field.type;
	let isMsg = false;
	let ttype: string | number;
	switch (type) {
		case "int32":
		case "sint32":
		case "sfixed32":
			type = "number";
			ttype = Type_Int32;
			break;
		case "enum":
		case "fixed32":
		case "uint32":
			type = "number";
			ttype = Type_Uint32;
		case "int64":
		case "uint64":
		case "sint64":
		case "double":
		case "float":
			type = "number";
			ttype = Type_Double;
			break;
		case "bool":
			type = "boolean";
			ttype = Type_Boolean;
			break;
		case "bytes":
			type = "Byte";
			ttype = Type_Bytes;
			break;
		case "fixed64":
		case "sfixed64":
			// 项目理论上不使用
			type = "Int64";
			break;
		case "message":
			type = field.type;
			isMsg = true;
			ttype = `"${type}"`;
			break;
		case "string":
			type = "string";
			ttype = Type_String;
			break;
		default:
			type = field.type;
			ttype = `"${type}"`;
			break;
	}
	if (field.rule == "repeated") { // 数组赋值
		return [type + "[]", isMsg, ttype, true];
	}
	return [type, isMsg, ttype, false];
}


function getVariable(field: pbjs.ProtoField, variables: string[]): FieldData 
{
	let comment = field.comment;// 调整protobuf.js代码 让其记录注释
	let fname = field.name;
	let def = "";//改成不赋值默认值，这样typescript->js能减少字符串输出，将默认值记入mMessageEncode，也方便传输时做判断
	let [fieldType, isMsg, tType, repeated] = field2type(field);
	if (field.rule == "required") { // 必选参数
		variables.push(`/**`);
		variables.push(` * ${comment}`);
		variables.push(` */`);
		variables.push(`${fname}: ${fieldType};`);
	} else {
		variables.push(`/**`);
		variables.push(` * 可选参数 ${comment}`);
		variables.push(` */`);
		variables.push(`${fname}?: ${fieldType};`);
	}
	return { fieldName: fname, fieldType: fieldType, isMsg: isMsg, tType: tType, repeated: repeated };
}

declare type FieldData = 
{
	/**
	 * 字段名称
	 */
	fieldName: string;
	/**
	 * 字段类型
	 */
	fieldType: string;
	/**
	 * 是否为消息类型
	 */
	isMsg: boolean;
	/**
	 * 用于注册typeScript的类型
	 */
	tType: number | string;
	/**
	 * 是否为repeated
	 */
	repeated: boolean;
}

function parseProto(proto: string, gcfg?: GlobalCfg) 
{
	let url: string = gcfg ? gcfg.url : "";
	url = url || "[文本框中，复制粘贴]";
	let cprefix = gcfg ? gcfg.cprefix || "" : "";
	let serviceName = gcfg ? gcfg.serviceName : null;
	let PBMsgDictName = gcfg ? gcfg.pBMsgDictName : null;
	let p = pbjs.DotProto.Parser.parse(proto);
	let options = p.options;
	let fcpath: string = options[Options.ClientPath];
	let fcmodule: string = options[Options.ClientModule];
	let service: string = options[Options.ServiceName];
	let now = new Date().format("yyyy-MM-dd HH:mm:ss");
	let idx = 0;
	let hasService = false;
	let msgEncDict: { [index: string]: string } = {};
	var msgTypeNameMap = {};
	// 客户端和服务端的Service收发数组
	let cSends: string[] = [], cRecvs: string[] = [], cRegs: string[] = [], sSends: string[] = [], sRecvs: string[] = [], sRegs: string[] = [];
	
	let pbMsgDictTemplate = new PBMsgDictTemplate();
	let pbMsgFilePath: string;
	let pbKeysDict: Map<string, string> = new Map();
	if (PBMsgDictName) {
		pbMsgFilePath = path.join(cprefix, PBMsgDictName);
		pbKeysDict = pbMsgDictTemplate.getKeys(pbMsgFilePath);
	}

	for (let msg of p.messages) {
		if (msg.name.indexOf('S2C') > 0 || msg.name.indexOf('C2S') > 0) {
			var key = msg.name.substring(0, msg.name.length - 8);
			if (msgTypeNameMap[key]) {
				msgTypeNameMap[key] = msgTypeNameMap[key] + 1;
			}
			else {
				msgTypeNameMap[key] = 1;
			}
		}
	}
	for (let msg of p.messages) {
		// 所有的变量
		// 一行一个
		let variables: string[] = [];
		let messageEncodes: string[] = [];
		// 如果有消息体的同名 Option，覆盖
		let options = msg.options;
		let cpath: string = options[Options.ClientPath] || fcpath;
		let cmodule: string = options[Options.ClientModule] || fcmodule;
		let limit: any = options[Options.Limit];
		var comment = msg.comment;
		let cmds: number[] = [];
		let requireds: FieldData[] = [];
		let repeateds: FieldData[] = [];
		let optionals: FieldData[] = [];

		if (comment.indexOf('=') > 0) {
			if (comment.indexOf('|') > 0) {
				cmds = [comment.split('|')[0].split('=')[1]];
			}
			else {
				cmds = [comment.split('=')[1]];
			}
		}

		let fieldids = [0];
		let paramCount = 0;
		for (let field of msg.fields) 
		{
			let rule: string = field.rule;
			let fnumber = type2number[field.type];
			let MType = "";
			let def = field.options.default;
			if (def !== undefined) {
				let t = typeof def;
				if (t === "string") {
					def = `"${def}"`;
				} else if (t === "object") {//不支持对象类型默认值
					def = undefined;
				}
			}
			if (fnumber == undefined) {// message类型不支持默认值
				fnumber = 11;// message
				MType = `, "${field.type}"`;
			} else if (def !== undefined) {
				MType = `, , ${def}`;
			}
			fieldids.push(~~field.id);
			messageEncodes.push(`${field.id}: ["${field.name}", ${rule2number[rule]}, ${fnumber}${MType}]`);
			let data = getVariable(field, variables);
			if (rule == "required") {
				requireds.push(data);
			} else if (rule == "repeated") {
				repeateds.push(data);
			} else {
				optionals.push(data);
			}
		}

		fieldids.sort((a,b)=>{return a - b});
		for (let i=0, l=fieldids.length-1; i < l; i++){
			let a = fieldids[i];
			let b = fieldids[i+1];
			if (b - a != 1){
				alert(`${msg.name} 中 value 可能定义有错误。顺序为从1开始，或未连续。此错误可能导致前后端协议值解析不正确`);
				break;
			}
		}

		// 根据CMD 生成通信代码
		// 生成代码
		let className: string = msg.name;
		let classNameFileName = path.join(cprefix, cpath, className + ".ts");
		let isExistsSync = fs.existsSync(classNameFileName);
		if ((!isExistsSync && pbKeysDict.has(className))){
			alert(`少年，${className} 已经在其他功能模块中存在了，请更换一个名字吧，当前文件未保存`);
			continue;
		}
		if (msgEncDict[className]){
			alert(`少年，好粗心啊，当前wiki地址中 ${className} 重名了，当前文件未保存`);
			continue;
		}

		msgEncDict[className] = `{ ${messageEncodes.join(", ")} }`;
		var typeVar = '';
		if (className.indexOf('S2C') > 0 || className.indexOf('C2S') > 0) {
			typeVar = className.substr(-7, 3);
		}
		let type = typeVar;
		let handlerName = className[0].toLowerCase() + className.substring(1, className.length - 4);
		let p = requireds.concat(repeateds, optionals);
		hasService = true;
		if (type == "C2S") { // client to server
			makeCSendFunction(p, className, handlerName, cSends, cmds[0], limit);
			// makeReciveFunc(className, handlerName + s2c, sRegs, sRecvs, cmds, p, true);
		} else if (type == "S2C") { // server to client
			makeReciveFunc(className, handlerName, cRegs, cRecvs, cmds, p, false);
			// if (messageEncodes.length == 1){
			// 	if (messageEncodes[0].indexOf("Code") == -1){
			// 		alert(`警告：${className} 协议接受服务器一个参数，可能导致前端解析不正确，随便增加一个参数吧`);
			// 	}
			// }
		} else {
			makeCSendFunction(p, className, handlerName + "_C2S", cSends, cmds[0], limit);
			makeReciveFunc(className, handlerName + "_S2C", cRegs, cRecvs, cmds, p, false);
		}
		let fleng = msg.fields.length;
		let clientCode = getClientCode(now, url, className, cmodule, variables);
		if (cprefix) {let outpath = path.join(cprefix, cpath, className + ".ts");
			let cdir = path.join(cprefix, cpath);
			let out = writeFile(className + ".ts", cdir, clientCode);
			if (out) {
				log(`<font color="#0c0">成功生成代码，${out}</font>`);
			}
		}
	}
	
	//处理PBMsgDict文件
	if (pbMsgFilePath) {
		let crout = pbMsgDictTemplate.addToFile(pbMsgFilePath, msgEncDict, false);
		let out = writeFile(PBMsgDictName, cprefix, crout);
		if (out) {
			log(`<font color="#0c0">生成PBMsgDict代码成功，${out}</font>`);
		}
	}

	if (service && hasService) {
		//预处理Service
		//检查是否有客户端Service文件
		let cdir = path.join(cprefix, fcpath);
		let sfileName = service + ".ts";
		let cpath = path.join(cdir, sfileName);
		let ccode = getCServiceCode(now, url, service, fcmodule, cSends, cRecvs, cRegs, getManualCodeInfo(cpath));
		//检查是否有服务端Service文件
		// 创建客户端Service
		if (cprefix) {
			let out = writeFile(sfileName, cdir, ccode);
			if (out) {
				log(`<font color="#0c0">${service}代码成功生成，${out}</font>`);
			}
		}
		// 创建ServiceName常量文件
		if (serviceName) {
			let ctemp = new ServiceNameTemplate();
			let cnPath = path.join(cprefix, serviceName);
			let code = ctemp.addToFile(cnPath, service);
			let out = writeFile(serviceName, cprefix, code);
			if (out) {
				log(`<font color="#0c0">ServiceName代码成功生成，${out}</font>`);
			}
		}
	}
}

function makeCSendFunction(fnames: FieldData[], className: string, handlerName: string, sends: string[], cmd: number, limit?: number) {
	let len = fnames.length;
	let time = limit ? ", "+ limit : "";
	if (len == 0) { //没有参数
		sends.push(`public ${handlerName}() {`);
		sends.push(`\tthis.send(${cmd}, null, ""${time})`);
	} else {
		sends.push(`public ${handlerName}(_${className}: ${className}) {`);
		sends.push(`\tthis.send(${cmd}, _${className}, "${className}"${time});`);
	}
	sends.push(`}`)
}

function makeReciveFunc(className: string, handlerName: string, regs: string[], recvs: string[], cmds: number[], fnames: FieldData[], isServer: boolean) {
	let len = fnames.length;
	let strCMD = cmds.join(",");
	var p: FieldData;
	if (len == 0) {//为null
		regs.push(`self.regMsg(${Type_Null}, ${strCMD});`);
	}else {
		regs.push(`self.regMsg("${className}", ${strCMD});`);
	}
	regs.push(`self.regHandler(self.${handlerName}, ${strCMD});`);
	let pp = isServer ? "async " : "";
	recvs.push(`protected ${handlerName} = ${pp}(data:NetData) => {`);
	recvs.push(`\tlet msg:${className} = <${className}>data.data;`);
	recvs.push(`/*|$${handlerName}|*/`);
	recvs.push(`}`);
}

function writeFile(fname: string, directory: string, data: string, corver = true): string {
	let outpath = path.join(directory, fname);
	try {
		FsExtra.writeFileSync(outpath, data, corver);
	} catch (e) {
		error(`写入文件时失败，路径："${directory}"，文件名："${fname}"，错误信息：${e.message}\n${e.stack}`);
		return null;
	}
	return outpath;
}

function getClientCode(createTime: string, url: string, className: string, module: string, variables: string[]) 
{	
	// let clientTemplatePath = "template/msg.template"//path.join("D:/workspace/bitbucket/miniprojects/h5workspace/ProtoTools/dist/", "template", "msg.template");
	// if (fs.existsSync(clientTemplatePath)) {
		//读取文件内容
		let content = clientMsgTemplate;//fs.readFileSync(clientTemplatePath, "utf8");
		
		content = content.replace(/\{path\}/g, url);
		content = content.replace(/\{createTime\}/g, createTime);
		content = content.replace(/\{className\}/g, className);
		content = content.replace(/\{vars\}/g, variables.join(`\n		`));

		return content;
	// }
	// return "";
}

function getCServiceCode(createTime: string, url: string, className: string, module: string, sends: string[], recvs: string[], regs: string[], cinfo: { [index: string]: string }) 
{
	// let clientTemplatePath = "template/client.template";
	// if (fs.existsSync(clientTemplatePath)) {
		//读取文件内容
		let content = clientServiceTemplate;//fs.readFileSync(clientTemplatePath, "utf8");

		let area1 = genManualAreaCode("$area1", cinfo);
		let area2 = genManualAreaCode("$area2", cinfo);
		let area3 = genManualAreaCode("$area3", cinfo);
		let onRegister = genManualAreaCode("$onRegister", cinfo);

		content = content.replace(/\{path\}/g, url);
		content = content.replace(/\{createTime\}/g, createTime);
		content = content.replace(/\{className\}/g, className);
		content = content.replace(/\{c2sfunctions\}/g, sends.join(`\n\t\t`));
		content = content.replace(/\{s2cfunctions\}/g, parseRecvs(recvs, cinfo));
		content = content.replace(/\{onRegister\}/g, onRegister);
		content = content.replace(/\{regMsg\}/g, regs.join(`\n\t\t\t`));
		content = content.replace(/\{area1\}/g, area1);
		content = content.replace(/\{area2\}/g, area2);
		content = content.replace(/\{area3\}/g, area3);

		return content;
	// }
	// return "the template file load failed";
}

function parseRecvs(recvs: string[], cinfo) 
{
	return recvs.map(recv => {
		return recv.replace(/[/][*][|]([$_a-zA-Z0-9]+)[|][*][/]/, (rep, hander) => {
			console.log(rep, hander);
			return genManualAreaCode(hander, cinfo);
		})
	}).join(`\n\t\t`);
}

/**
 * 获取http数据
 */
function getHttpData(url: string, gcfg?: GlobalCfg) 
{
	let promise = new Promise((resolve, reject) => {
		let req = http.get(url, res => {
			let size = 0;
			let chunks: Buffer[] = [];
			res.on("data", (chunk: Buffer) => {
				size += chunk.length;
				chunks.push(chunk);
			});
			res.on("end", () => {
				let data = Buffer.concat(chunks, size);
				resolve({ content: data.toString("utf8"), gcfg: gcfg, url: url });
			});
		});
		req.on("error", (err) => {
			reject(err.message);
		});
	});
	return promise;
}

/**
 * 提取Proto数据
 */
function getProtoData(data: { content: string, gcfg?: GlobalCfg, url: string }) 
{
	// let content = escapeHTML(data.content);
	// 从HTML流中截取 message {} 或者 option (xxxx) = "xxxx" 这样的数据
	// let reg: RegExp = /<pre class="code">([^>]*?message[ ]+[A-Z][a-zA-Z0-9_$]*[ ]*[{][^]*?[}][^]*?|[^>]*?option[ ]+[(]\w+[)][ ]*=[ ]*".*?";[^]*?)<\/pre>/mg;
	// let proto = "";
	// while (true) {
	// 	let result = reg.exec(content);
	// 	if (result) {
	// 		proto += result[1] + "\n";
	// 	} else {
	// 		break;
	// 	}
	// }
	// proto =proto.substring(proto.indexOf("option "));
	$g("txtProto").value = data.content;
	parseProto(data.content, data.gcfg);
}

const escChars = { "&lt;": "<", "&gt;": ">", "&quot;": "\"", "&apos;": "\'", "&amp;": "&", "&nbsp;": " ", "&#x000A;": "\n", "&#039;": "\'", "\\n": "\n"};
function escapeHTML(content: string) {
	return content.replace(/&lt;|&gt;|&quot;|&apos;|&amp;|&nbsp;|&#x000A;/g, substring => {
		return escChars[substring];
	});
}

function escapeString(content: string) {
	return content.replace(/&lt;|&gt;|&quot;|&apos;|&amp;|&nbsp;|&#x000A;|&#039;|\\n/g, substring => {
		return escChars[substring];
	});
}

function getProtoFromHttp(url: string, gcfg?: GlobalCfg) 
{
	let promise = getHttpData(url, gcfg);
	promise.then(getProtoData, error).catch(error);
}

function getUrlErrorCode(data) {
	let clientConst = [];
	//先尝试读取客户端的常量文件
	const FileName = "MsgConst.ts";
	let gcfg = data.gcfg;
	let tsDict:{[title:string]: {code: string, desc: string}[]} = {};
	let url;
	let titles = [];
	let content = escapeHTML(data.content);
	let reg = /<h\d+\sclass="sectionedit\d+"\sid=".*?">(.*?)[|](.*?)<\/h\d+>\s+<div\sclass="level\d+">\s+<pre\sclass="code">([\s\S]*?)<\/pre>/mg;
	let titleComments = {};
	let retdata: {[code: string]: string} = {};
	const tempChars = { "\\n": "\n"};

	while (true) {
		let result = reg.exec(content);
		if (result) {
			let [,titleComment, title, code] = result;
			if (code) {
				if (!~titles.indexOf(title)) {
					titles.push(title);
					titleComments[title] = titleComment;
					let codes:{code: string, desc: string}[] = [];
					tsDict[title] = codes;
					let data = code.replace(/\r/g, "").split(/\n/g);
					data.forEach(str=>{
						let temp = str.split(/\s+/);
						let c = temp[0];
						let d = escapeString(str.substring(c.length + 1));
						if (c){
							if (isNaN(Number(c))){
								error(`【定义错误】${title}中Code码key[${c}]必须是数字`)
							}
							else if (retdata[c]) {
								error(`【定义重复】${title}中Code码key[${c}]已经定义`)
							}else{
								retdata[c] = d;
								codes.push({code: c, desc: d});
							}
						}
					})
				}
			}
		}
		else {
			break;
		}
	}
	titles.forEach(title => {
		let map = tsDict[title];
		if (map.length > 0) {
			clientConst.push(`/** ${titleComments[title]} */`);
			clientConst.push(`const enum $${title} {`);
			map.forEach(item => {
				clientConst.push(`\t/** ${item.desc} */`);
				clientConst.push(`\t$${item.code} = ${item.code},`);
			});
			clientConst.push(`}`);
		}
	});
	let clientCode = clientConst.join("\n");
	if (gcfg.cprefix) {
		url = path.join(gcfg.cprefix, FileName);
		fs.writeFileSync(url, clientCode);
	} else {
		error(`没有【客户端项目根目录(src)】地址，保存客户端代码失败`);
	}
	console.log(clientCode);
	log(`<font color="#0c0">开始生成errorcode</font>`);
	let txtLocalCodePath = cookieForPath.setPathCookie("txtLocalCodePath", false, false);
	if (txtLocalCodePath){
		let codedata = JSON.stringify(retdata);
		url = path.join(txtLocalCodePath, "code.json");
		fs.writeFileSync(url, codedata);
		// url = path.join(txtLocalCodePath, "code.js");
		// fs.writeFileSync(url, "var $errorcode =" + codedata);
	}
	else {
		error("保存code.js数据文件失败，没有填写存储地址");
	}
	log(`<font color="#0c0">生成errorcode结束</font>`);
}

const cookieForPath = new CookieForPath("protocal_history_");

ready(() => 
{
	cookieForPath.getPathCookie("txtClientPath");
	cookieForPath.getPathCookie("txtServerServiceNamePath");
	cookieForPath.getPathCookie("txtPBMsgDictName");
	cookieForPath.getPathCookie("txtUrl");
	cookieForPath.getPathCookie("txtWikiCodePath");
	cookieForPath.getPathCookie("txtLocalCodePath");
	$g("btnGen").addEventListener("click", (ev) => {
		let cPath = cookieForPath.setPathCookie("txtClientPath");
		let serviceName = cookieForPath.setPathCookie("txtServerServiceNamePath", false, false) || "ServiceName.ts";
		let pBMsgDictName = cookieForPath.setPathCookie("txtPBMsgDictName", false, false) || "PBMsgDict.ts";
		let url: string = cookieForPath.setPathCookie("txtUrl", false, false);
		url = url ? url.trim() : "";
		if (!url){
			error("请输入wiki protocol地址");
			return;
		}
		// 检查url路径
		let gcfg: GlobalCfg = { cprefix: cPath, url: url, serviceName: serviceName, pBMsgDictName: pBMsgDictName };
		getProtoFromHttp(url, gcfg);
	});
	$g("btnCode").addEventListener("click", (ev) => {
		let cPath = cookieForPath.setPathCookie("txtClientPath");
		let txtWikiCodePath = cookieForPath.setPathCookie("txtWikiCodePath", false, false);
		let txtLocalCodePath = cookieForPath.setPathCookie("txtLocalCodePath", false, false);
		if (!txtWikiCodePath) {
			error(`code wiki url 不能为空`);
			return;
		}
		if (!txtLocalCodePath) {
			error(`code 存储 url 不能为空`);
			return ;
		}
		// if (!cPath) {
		// 	error(`code 存储 url 为空， 没有保存客户端代码文件`);
		// }
		// 检查url路径
		let gcfg: GlobalCfg = { cprefix: cPath };
		let urlErrorPromise = getHttpData(txtWikiCodePath, gcfg);
		urlErrorPromise.then(getUrlErrorCode, error).catch(error);
	})
});

/**
 * 创建一个代码区
 */
// function createContent(parent: HTMLElement, filename: string, idx: number, ccode: string) {
// 	let pane = document.createElement("div");
// 	pane.style.border = "#000 solid 1px";
// 	let idCopyClient = "btnCopyClient" + idx;
// 	let template = `<div>${filename}</div>
//     <div style="width:50%;height:50%;float:left;background:#eef">
//         客户端代码：<input type="button" value="复制客户端代码" id="${idCopyClient}" />
//         <textarea style="width:100%;height:200px;border:#ccf solid 1px;background:#eee" contenteditable="false">${ccode}</textarea>
//     </div>`
// 	pane.innerHTML = template;
// 	parent.appendChild(pane);
// 	$g(idCopyClient).addEventListener("click", e => {
// 		clipboard.writeText(ccode);
// 	});
// }

/** 手写代码的默认提示 */
const ManualCodeDefaultComment = 
{
	/** 这里手写类上方的内容 */
	$area1: "@这里手写的内容，请在注释begin 和 end 之间编写自己的代码，不可更改此注释@",
	/** 这里手写类里面的内容 */
	$area2: "@这里手写的内容，请在注释begin 和 end 之间编写自己的代码，不可更改此注释@",
	/** 这里手写类下的内容 */
	$area3: "@这里手写的内容，请在注释begin 和 end 之间编写自己的代码，不可更改此注释@",
	/** 这里手写onRegister中内容 */
	$onRegister: "@这里手写的内容，请在注释begin 和 end 之间编写自己的代码，不可更改此注释@",
	/** 这里手写方法中的内容 */
	$handler: "@这里手写的内容，请在注释begin 和 end 之间编写自己的代码，不可更改此注释@",
}

/**
 * 生成手动代码区域的文本
 */
function genManualAreaCode(key: string, cinfo: { [index: string]: string }) 
{
	let manual = cinfo[key];
	let tab = "\t";
	let btab = "";
	let desc = ManualCodeDefaultComment[key] || ManualCodeDefaultComment.$handler;
	if (key in ManualCodeDefaultComment) {
		if (key == "$onRegister") tab = "\t\t\t";
		else if (key == "$area2") tab = "\t\t";
	} else {
		tab = "\t\t\t";
		btab = "\t";
	}
	return `${btab}/*-*begin ${key} ${desc}*-*/
${tab}${manual ? manual : ""}
${tab}/*-*end ${key}*-*/`
}

/**
 * 获取手动写的代码信息
 */
function getManualCodeInfo(file: string) {
	let dict: { [index: string]: string } = {};
	if (file && fs.existsSync(file)) {
		//读取文件内容
		let content = fs.readFileSync(file, "utf8");
		//找到手写内容
		let reg = /[/][*]-[*]begin[ ]([$_a-zA-Z0-9]+)[ ](.*?)[*]-[*][/]([^]*?)[/][*]-[*]end[ ]([$_a-zA-Z0-9]+)[*]-[*][/]/g
		console.log("*****************************start");
		while (true) {
			let result = reg.exec(content);
			if (result) {
				let key = result[1];
				let manual = result[3].trim();
				console.log(key, manual);
				if (!manual) {//没有注释
					continue;
				}
				dict[key] = manual;
			} else {
				break;
			}
		}
	}
	console.log("=================================");
	console.log(dict);
	console.log("=================================");
	return dict;
}

////////////////////////template///////////////////

var clientServiceTemplate = `
/**
 * 来源： {path}
 */
module giant.chuanqi 
{
	{area1}
	
	export class {className} extends Service 
	{
		constructor() 
		{
			super(ServiceName.{className});
		}

		onRegister() 
		{
			super.onRegister();

			let self = this;
			
			{regMsg}

			{onRegister}
		}

		{c2sfunctions}

		{s2cfunctions}
		
		{area2}
		
	}
	
	{area3}
	
}
`;

var clientMsgTemplate = `
/**
 * 源地址： {path}
 */
module giant.chuanqi 
{
	export interface {className}
	{
		
		{vars}
		
	}
}
`;